"""Test cases for the isoduration module."""

from datetime import date, datetime, timedelta

import pytest

from isodate import (
    D_ALT_EXT,
    D_DEFAULT,
    D_WEEK,
    Duration,
    ISO8601Error,
    duration_isoformat,
    parse_duration,
)

# the following list contains tuples of ISO duration strings and the expected
# result from the parse_duration method. A result of None means an ISO8601Error
# is expected.
PARSE_TEST_CASES = (
    ("P18Y9M4DT11H9M8S", Duration(4, 8, 0, 0, 9, 11, 0, 9, 18), D_DEFAULT, None),
    ("P2W", timedelta(weeks=2), D_WEEK, None),
    ("P3Y6M4DT12H30M5S", Duration(4, 5, 0, 0, 30, 12, 0, 6, 3), D_DEFAULT, None),
    ("P23DT23H", timedelta(hours=23, days=23), D_DEFAULT, None),
    ("P4Y", Duration(years=4), D_DEFAULT, None),
    ("P1M", Duration(months=1), D_DEFAULT, None),
    ("PT1M", timedelta(minutes=1), D_DEFAULT, None),
    ("P0.5Y", Duration(years=0.5), D_DEFAULT, None),
    ("PT36H", timedelta(hours=36), D_DEFAULT, "P1DT12H"),
    ("P1DT12H", timedelta(days=1, hours=12), D_DEFAULT, None),
    ("+P11D", timedelta(days=11), D_DEFAULT, "P11D"),
    ("-P2W", timedelta(weeks=-2), D_WEEK, None),
    ("-P2.2W", timedelta(weeks=-2.2), D_DEFAULT, "-P15DT9H36M"),
    ("P1DT2H3M4S", timedelta(days=1, hours=2, minutes=3, seconds=4), D_DEFAULT, None),
    ("P1DT2H3M", timedelta(days=1, hours=2, minutes=3), D_DEFAULT, None),
    ("P1DT2H", timedelta(days=1, hours=2), D_DEFAULT, None),
    ("PT2H", timedelta(hours=2), D_DEFAULT, None),
    ("PT2.3H", timedelta(hours=2.3), D_DEFAULT, "PT2H18M"),
    ("PT2H3M4S", timedelta(hours=2, minutes=3, seconds=4), D_DEFAULT, None),
    ("PT3M4S", timedelta(minutes=3, seconds=4), D_DEFAULT, None),
    ("PT22S", timedelta(seconds=22), D_DEFAULT, None),
    ("PT22.22S", timedelta(seconds=22.22), "PT%S.%fS", "PT22.220000S"),
    ("-P2Y", Duration(years=-2), D_DEFAULT, None),
    ("-P3Y6M4DT12H30M5S", Duration(-4, -5, 0, 0, -30, -12, 0, -6, -3), D_DEFAULT, None),
    (
        "-P1DT2H3M4S",
        timedelta(days=-1, hours=-2, minutes=-3, seconds=-4),
        D_DEFAULT,
        None,
    ),
    # alternative format
    ("P0018-09-04T11:09:08", Duration(4, 8, 0, 0, 9, 11, 0, 9, 18), D_ALT_EXT, None),
    # 'PT000022.22', timedelta(seconds=22.22),
)

#                       d1                    d2           '+', '-', '>'
# A list of test cases to test addition and subtraction between datetime and
# Duration objects.
# each tuple contains 2 duration strings, and a result string for addition and
# one for subtraction. The last value says, if the first duration is greater
# than the second.
MATH_TEST_CASES = (
    (
        "P5Y7M1DT9H45M16.72S",
        "PT27M24.68S",
        "P5Y7M1DT10H12M41.4S",
        "P5Y7M1DT9H17M52.04S",
        None,
    ),
    ("PT28M12.73S", "PT56M29.92S", "PT1H24M42.65S", "-PT28M17.19S", False),
    (
        "P3Y7M23DT5H25M0.33S",
        "PT1H1.95S",
        "P3Y7M23DT6H25M2.28S",
        "P3Y7M23DT4H24M58.38S",
        None,
    ),
    (
        "PT1H1.95S",
        "P3Y7M23DT5H25M0.33S",
        "P3Y7M23DT6H25M2.28S",
        "-P3Y7M23DT4H24M58.38S",
        None,
    ),
    ("P1332DT55M0.33S", "PT1H1.95S", "P1332DT1H55M2.28S", "P1331DT23H54M58.38S", True),
    (
        "PT1H1.95S",
        "P1332DT55M0.33S",
        "P1332DT1H55M2.28S",
        "-P1331DT23H54M58.38S",
        False,
    ),
)


# A list of test cases to test addition and subtraction of date/datetime
# and Duration objects. They are tested against the results of an
# equal long timedelta duration.
DATE_TEST_CASES = (
    (
        date(2008, 2, 29),
        timedelta(days=10, hours=12, minutes=20),
        Duration(days=10, hours=12, minutes=20),
    ),
    (
        date(2008, 1, 31),
        timedelta(days=10, hours=12, minutes=20),
        Duration(days=10, hours=12, minutes=20),
    ),
    (
        datetime(2008, 2, 29),
        timedelta(days=10, hours=12, minutes=20),
        Duration(days=10, hours=12, minutes=20),
    ),
    (
        datetime(2008, 1, 31),
        timedelta(days=10, hours=12, minutes=20),
        Duration(days=10, hours=12, minutes=20),
    ),
    (
        datetime(2008, 4, 21),
        timedelta(days=10, hours=12, minutes=20),
        Duration(days=10, hours=12, minutes=20),
    ),
    (
        datetime(2008, 5, 5),
        timedelta(days=10, hours=12, minutes=20),
        Duration(days=10, hours=12, minutes=20),
    ),
    (datetime(2000, 1, 1), timedelta(hours=-33), Duration(hours=-33)),
    (
        datetime(2008, 5, 5),
        Duration(years=1, months=1, days=10, hours=12, minutes=20),
        Duration(months=13, days=10, hours=12, minutes=20),
    ),
    (
        datetime(2000, 3, 30),
        Duration(years=1, months=1, days=10, hours=12, minutes=20),
        Duration(months=13, days=10, hours=12, minutes=20),
    ),
)

# A list of test cases of addition of date/datetime and Duration. The results
# are compared against a given expected result.
DATE_CALC_TEST_CASES = (
    (date(2000, 2, 1), Duration(years=1, months=1), date(2001, 3, 1)),
    (date(2000, 2, 29), Duration(years=1, months=1), date(2001, 3, 29)),
    (date(2000, 2, 29), Duration(years=1), date(2001, 2, 28)),
    (date(1996, 2, 29), Duration(years=4), date(2000, 2, 29)),
    (date(2096, 2, 29), Duration(years=4), date(2100, 2, 28)),
    (date(2000, 2, 1), Duration(years=-1, months=-1), date(1999, 1, 1)),
    (date(2000, 2, 29), Duration(years=-1, months=-1), date(1999, 1, 29)),
    (date(2000, 2, 1), Duration(years=1, months=1, days=1), date(2001, 3, 2)),
    (date(2000, 2, 29), Duration(years=1, months=1, days=1), date(2001, 3, 30)),
    (date(2000, 2, 29), Duration(years=1, days=1), date(2001, 3, 1)),
    (date(1996, 2, 29), Duration(years=4, days=1), date(2000, 3, 1)),
    (date(2096, 2, 29), Duration(years=4, days=1), date(2100, 3, 1)),
    (date(2000, 2, 1), Duration(years=-1, months=-1, days=-1), date(1998, 12, 31)),
    (date(2000, 2, 29), Duration(years=-1, months=-1, days=-1), date(1999, 1, 28)),
    (date(2001, 4, 1), Duration(years=-1, months=-1, days=-1), date(2000, 2, 29)),
    (date(2000, 4, 1), Duration(years=-1, months=-1, days=-1), date(1999, 2, 28)),
    (
        Duration(years=1, months=2),
        Duration(years=0, months=0, days=1),
        Duration(years=1, months=2, days=1),
    ),
    (Duration(years=-1, months=-1, days=-1), date(2000, 4, 1), date(1999, 2, 28)),
    (Duration(years=1, months=1, weeks=5), date(2000, 1, 30), date(2001, 4, 4)),
    (parse_duration("P1Y1M5W"), date(2000, 1, 30), date(2001, 4, 4)),
    (parse_duration("P0.5Y"), date(2000, 1, 30), None),
    (
        Duration(years=1, months=1, hours=3),
        datetime(2000, 1, 30, 12, 15, 00),
        datetime(2001, 2, 28, 15, 15, 00),
    ),
    (
        parse_duration("P1Y1MT3H"),
        datetime(2000, 1, 30, 12, 15, 00),
        datetime(2001, 2, 28, 15, 15, 00),
    ),
    (
        Duration(years=1, months=2),
        timedelta(days=1),
        Duration(years=1, months=2, days=1),
    ),
    (
        timedelta(days=1),
        Duration(years=1, months=2),
        Duration(years=1, months=2, days=1),
    ),
    (datetime(2008, 1, 1, 0, 2), Duration(months=1), datetime(2008, 2, 1, 0, 2)),
    (
        datetime.strptime("200802", "%Y%M"),
        parse_duration("P1M"),
        datetime(2008, 2, 1, 0, 2),
    ),
    (datetime(2008, 2, 1), Duration(months=1), datetime(2008, 3, 1)),
    (datetime.strptime("200802", "%Y%m"), parse_duration("P1M"), datetime(2008, 3, 1)),
    # (date(2000, 1, 1),
    #  Duration(years=1.5),
    #  date(2001, 6, 1)),
    # (date(2000, 1, 1),
    #  Duration(years=1, months=1.5),
    #  date(2001, 2, 14)),
)

# A list of test cases of multiplications of durations
# are compared against a given expected result.
DATE_MUL_TEST_CASES = (
    (Duration(years=1, months=1), 3, Duration(years=3, months=3)),
    (Duration(years=1, months=1), -3, Duration(years=-3, months=-3)),
    (3, Duration(years=1, months=1), Duration(years=3, months=3)),
    (-3, Duration(years=1, months=1), Duration(years=-3, months=-3)),
    (5, Duration(years=2, minutes=40), Duration(years=10, hours=3, minutes=20)),
    (-5, Duration(years=2, minutes=40), Duration(years=-10, hours=-3, minutes=-20)),
    (7, Duration(years=1, months=2, weeks=40), Duration(years=8, months=2, weeks=280)),
)


def test_associative():
    """Adding 2 durations to a date is not associative."""
    days1 = Duration(days=1)
    months1 = Duration(months=1)
    start = date(2000, 3, 30)
    res1 = start + days1 + months1
    res2 = start + months1 + days1
    assert res1 != res2


def test_typeerror():
    """Test if TypError is raised with certain parameters."""
    with pytest.raises(TypeError):
        parse_duration(date(2000, 1, 1))
    with pytest.raises(TypeError):
        Duration(years=1) - date(2000, 1, 1)
    with pytest.raises(TypeError):
        "raise exc" - Duration(years=1)
    with pytest.raises(TypeError):
        Duration(years=1, months=1, weeks=5) + "raise exception"
    with pytest.raises(TypeError):
        "raise exception" + Duration(years=1, months=1, weeks=5)
    with pytest.raises(TypeError):
        Duration(years=1, months=1, weeks=5) * "raise exception"
    with pytest.raises(TypeError):
        "raise exception" * Duration(years=1, months=1, weeks=5)
    with pytest.raises(TypeError):
        Duration(years=1, months=1, weeks=5) * 3.14
    with pytest.raises(TypeError):
        3.14 * Duration(years=1, months=1, weeks=5)


def test_parseerror():
    """Test for unparseable duration string."""
    with pytest.raises(ISO8601Error):
        parse_duration("T10:10:10")


def test_repr():
    """Test __repr__ and __str__ for Duration objects."""
    dur = Duration(10, 10, years=10, months=10)
    assert "10 years, 10 months, 10 days, 0:00:10" == str(dur)
    assert "isodate.duration.Duration(10, 10, 0," " years=10, months=10)" == repr(dur)
    dur = Duration(months=0)
    assert "0:00:00" == str(dur)
    dur = Duration(months=1)
    assert "1 month, 0:00:00" == str(dur)


def test_hash():
    """Test __hash__ for Duration objects."""
    dur1 = Duration(10, 10, years=10, months=10)
    dur2 = Duration(9, 9, years=9, months=9)
    dur3 = Duration(10, 10, years=10, months=10)
    assert hash(dur1) != hash(dur2)
    assert id(dur1) != id(dur2)
    assert hash(dur1) == hash(dur3)
    assert id(dur1) != id(dur3)
    durSet = set()
    durSet.add(dur1)
    durSet.add(dur2)
    durSet.add(dur3)
    assert len(durSet) == 2


def test_neg():
    """Test __neg__ for Duration objects."""
    assert -Duration(0) == Duration(0)
    assert -Duration(years=1, months=1) == Duration(years=-1, months=-1)
    assert -Duration(years=1, months=1) == Duration(months=-13)
    assert -Duration(years=1) != timedelta(days=-365)
    assert -timedelta(days=365) != Duration(years=-1)
    # FIXME: this test fails in python 3... it seems like python3
    #        treats a == b the same b == a
    # assert -timedelta(days=10) != -Duration(days=10)


def test_format():
    """Test various other strftime combinations."""
    assert duration_isoformat(Duration(0)) == "P0D"
    assert duration_isoformat(-Duration(0)) == "P0D"
    assert duration_isoformat(Duration(seconds=10)) == "PT10S"
    assert duration_isoformat(Duration(years=-1, months=-1)) == "-P1Y1M"
    assert duration_isoformat(-Duration(years=1, months=1)) == "-P1Y1M"
    assert duration_isoformat(-Duration(years=-1, months=-1)) == "P1Y1M"
    assert duration_isoformat(-Duration(years=-1, months=-1)) == "P1Y1M"
    dur = Duration(years=3, months=7, days=23, hours=5, minutes=25, milliseconds=330)
    assert duration_isoformat(dur) == "P3Y7M23DT5H25M0.33S"
    assert duration_isoformat(-dur) == "-P3Y7M23DT5H25M0.33S"


def test_equal():
    """Test __eq__ and __ne__ methods."""
    assert Duration(years=1, months=1) == Duration(years=1, months=1)
    assert Duration(years=1, months=1) == Duration(months=13)
    assert Duration(years=1, months=2) != Duration(years=1, months=1)
    assert Duration(years=1, months=1) != Duration(months=14)
    assert Duration(years=1) != timedelta(days=365)
    assert (Duration(years=1, months=1) != Duration(years=1, months=1)) is False
    assert (Duration(years=1, months=1) != Duration(months=13)) is False
    assert (Duration(years=1, months=2) != Duration(years=1, months=1)) is True
    assert (Duration(years=1, months=1) != Duration(months=14)) is True
    assert (Duration(years=1) != timedelta(days=365)) is True
    assert Duration(days=1) == timedelta(days=1)
    # FIXME: this test fails in python 3... it seems like python3
    #        treats a != b the same b != a
    # assert timedelta(days=1) != Duration(days=1)


def test_totimedelta():
    """Test conversion form Duration to timedelta."""
    dur = Duration(years=1, months=2, days=10)
    assert dur.totimedelta(datetime(1998, 2, 25)) == timedelta(434)
    # leap year has one day more in february
    assert dur.totimedelta(datetime(2000, 2, 25)) == timedelta(435)
    dur = Duration(months=2)
    # march is longer than february, but april is shorter than
    # march (cause only one day difference compared to 2)
    assert dur.totimedelta(datetime(2000, 2, 25)) == timedelta(60)
    assert dur.totimedelta(datetime(2001, 2, 25)) == timedelta(59)
    assert dur.totimedelta(datetime(2001, 3, 25)) == timedelta(61)


@pytest.mark.parametrize(
    "durationstring, expectation, format, altstr",
    PARSE_TEST_CASES,
)
def test_parse(durationstring, expectation, format, altstr):
    """Parse an ISO duration string and compare it to the expected value."""
    result = parse_duration(durationstring)
    assert result == expectation


@pytest.mark.parametrize(
    "durationstring, expectation, format, altstr",
    PARSE_TEST_CASES,
)
def test_format_parse(durationstring, expectation, format, altstr):
    """Take duration/timedelta object and create ISO string from it.

    This is the reverse test to test_parse.
    """
    if altstr:
        assert duration_isoformat(expectation, format) == altstr
    else:
        # if durationstring == '-P2W':
        #     import pdb; pdb.set_trace()
        assert duration_isoformat(expectation, format) == durationstring


@pytest.mark.parametrize("dur1, dur2, resadd, ressub, resge", MATH_TEST_CASES)
def test_add(dur1, dur2, resadd, ressub, resge):
    dur1 = parse_duration(dur1)
    dur2 = parse_duration(dur2)
    resadd = parse_duration(resadd)
    assert dur1 + dur2 == resadd


@pytest.mark.parametrize("dur1, dur2, resadd, ressub, resge", MATH_TEST_CASES)
def test_sub(dur1, dur2, resadd, ressub, resge):
    """
    Test operator - (__sub__, __rsub__)
    """
    dur1 = parse_duration(dur1)
    dur2 = parse_duration(dur2)
    ressub = parse_duration(ressub)
    assert dur1 - dur2 == ressub


@pytest.mark.parametrize("dur1, dur2, resadd, ressub, resge", MATH_TEST_CASES)
def test_ge(dur1, dur2, resadd, ressub, resge):
    """Test operator > and <."""
    dur1 = parse_duration(dur1)
    dur2 = parse_duration(dur2)

    def dogetest():
        """Test greater than."""
        return dur1 > dur2

    def doletest():
        """Test less than."""
        return dur1 < dur2

    if resge is None:
        with pytest.raises(TypeError):
            dogetest()
        with pytest.raises(TypeError):
            doletest()
    else:
        assert dogetest() is resge
        assert doletest() is not resge


@pytest.mark.parametrize("start, tdelta, duration", DATE_TEST_CASES)
def test_add_date(start, tdelta, duration):
    assert start + tdelta == start + duration


@pytest.mark.parametrize("start, tdelta, duration", DATE_TEST_CASES)
def test_sub_date(start, tdelta, duration):
    assert start - tdelta == start - duration


@pytest.mark.parametrize("start, duration, expectation", DATE_CALC_TEST_CASES)
def test_calc_date(start, duration, expectation):
    """Test operator +."""
    if expectation is None:
        with pytest.raises(ValueError):
            start + duration
    else:
        assert start + duration == expectation


@pytest.mark.parametrize("operand1, operand2, expectation", DATE_MUL_TEST_CASES)
def test_mul_date(operand1, operand2, expectation):
    """Test operator *."""
    assert operand1 * operand2 == expectation



"""
Test cases for the isodatetime module.
"""

from datetime import datetime

import pytest

from isodate import (
    DATE_BAS_COMPLETE,
    DATE_BAS_ORD_COMPLETE,
    DATE_BAS_WEEK_COMPLETE,
    DATE_EXT_COMPLETE,
    DATE_EXT_ORD_COMPLETE,
    DATE_EXT_WEEK_COMPLETE,
    TIME_BAS_COMPLETE,
    TIME_BAS_MINUTE,
    TIME_EXT_COMPLETE,
    TIME_EXT_MINUTE,
    TZ_BAS,
    TZ_EXT,
    TZ_HOUR,
    UTC,
    FixedOffset,
    ISO8601Error,
    datetime_isoformat,
    parse_datetime,
)

# the following list contains tuples of ISO datetime strings and the expected
# result from the parse_datetime method. A result of None means an ISO8601Error
# is expected.
TEST_CASES = [
    (
        "19850412T1015",
        datetime(1985, 4, 12, 10, 15),
        DATE_BAS_COMPLETE + "T" + TIME_BAS_MINUTE,
        "19850412T1015",
    ),
    (
        "1985-04-12T10:15",
        datetime(1985, 4, 12, 10, 15),
        DATE_EXT_COMPLETE + "T" + TIME_EXT_MINUTE,
        "1985-04-12T10:15",
    ),
    (
        "1985102T1015Z",
        datetime(1985, 4, 12, 10, 15, tzinfo=UTC),
        DATE_BAS_ORD_COMPLETE + "T" + TIME_BAS_MINUTE + TZ_BAS,
        "1985102T1015Z",
    ),
    (
        "1985-102T10:15Z",
        datetime(1985, 4, 12, 10, 15, tzinfo=UTC),
        DATE_EXT_ORD_COMPLETE + "T" + TIME_EXT_MINUTE + TZ_EXT,
        "1985-102T10:15Z",
    ),
    (
        "1985W155T1015+0400",
        datetime(1985, 4, 12, 10, 15, tzinfo=FixedOffset(4, 0, "+0400")),
        DATE_BAS_WEEK_COMPLETE + "T" + TIME_BAS_MINUTE + TZ_BAS,
        "1985W155T1015+0400",
    ),
    (
        "1985-W15-5T10:15+04",
        datetime(
            1985,
            4,
            12,
            10,
            15,
            tzinfo=FixedOffset(4, 0, "+0400"),
        ),
        DATE_EXT_WEEK_COMPLETE + "T" + TIME_EXT_MINUTE + TZ_HOUR,
        "1985-W15-5T10:15+04",
    ),
    (
        "1985-W15-5T10:15-0430",
        datetime(
            1985,
            4,
            12,
            10,
            15,
            tzinfo=FixedOffset(-4, -30, "-0430"),
        ),
        DATE_EXT_WEEK_COMPLETE + "T" + TIME_EXT_MINUTE + TZ_BAS,
        "1985-W15-5T10:15-0430",
    ),
    (
        "1985-W15-5T10:15+04:45",
        datetime(
            1985,
            4,
            12,
            10,
            15,
            tzinfo=FixedOffset(4, 45, "+04:45"),
        ),
        DATE_EXT_WEEK_COMPLETE + "T" + TIME_EXT_MINUTE + TZ_EXT,
        "1985-W15-5T10:15+04:45",
    ),
    (
        "20110410T101225.123000Z",
        datetime(2011, 4, 10, 10, 12, 25, 123000, tzinfo=UTC),
        DATE_BAS_COMPLETE + "T" + TIME_BAS_COMPLETE + ".%f" + TZ_BAS,
        "20110410T101225.123000Z",
    ),
    (
        "2012-10-12T08:29:46.069178Z",
        datetime(2012, 10, 12, 8, 29, 46, 69178, tzinfo=UTC),
        DATE_EXT_COMPLETE + "T" + TIME_EXT_COMPLETE + ".%f" + TZ_BAS,
        "2012-10-12T08:29:46.069178Z",
    ),
    (
        "2012-10-12T08:29:46.691780Z",
        datetime(2012, 10, 12, 8, 29, 46, 691780, tzinfo=UTC),
        DATE_EXT_COMPLETE + "T" + TIME_EXT_COMPLETE + ".%f" + TZ_BAS,
        "2012-10-12T08:29:46.691780Z",
    ),
    (
        "2012-10-30T08:55:22.1234567Z",
        datetime(2012, 10, 30, 8, 55, 22, 123456, tzinfo=UTC),
        DATE_EXT_COMPLETE + "T" + TIME_EXT_COMPLETE + ".%f" + TZ_BAS,
        "2012-10-30T08:55:22.123456Z",
    ),
    (
        "2012-10-30T08:55:22.1234561Z",
        datetime(2012, 10, 30, 8, 55, 22, 123456, tzinfo=UTC),
        DATE_EXT_COMPLETE + "T" + TIME_EXT_COMPLETE + ".%f" + TZ_BAS,
        "2012-10-30T08:55:22.123456Z",
    ),
    (
        "2014-08-18 14:55:22.123456Z",
        None,
        DATE_EXT_COMPLETE + "T" + TIME_EXT_COMPLETE + ".%f" + TZ_BAS,
        "2014-08-18T14:55:22.123456Z",
    ),
]


@pytest.mark.parametrize("datetimestring, expected, format, output", TEST_CASES)
def test_parse(datetimestring, expected, format, output):
    """
    Parse an ISO datetime string and compare it to the expected value.
    """
    if expected is None:
        with pytest.raises(ISO8601Error):
            parse_datetime(datetimestring)
    else:
        assert parse_datetime(datetimestring) == expected


@pytest.mark.parametrize("datetimestring, expected, format, output", TEST_CASES)
def test_format(datetimestring, expected, format, output):
    """
    Take datetime object and create ISO string from it.
    This is the reverse test to test_parse.
    """
    if expected is None:
        with pytest.raises(AttributeError):
            datetime_isoformat(expected, format)
    else:
        assert datetime_isoformat(expected, format) == output



import pickle

import isodate


def test_pickle_datetime():
    """Parse an ISO datetime string and compare it to the expected value."""
    dti = isodate.parse_datetime("2012-10-26T09:33+00:00")
    for proto in range(0, pickle.HIGHEST_PROTOCOL + 1):
        pikl = pickle.dumps(dti, proto)
        assert dti == pickle.loads(pikl), "pickle proto %d failed" % proto


def test_pickle_duration():
    """Pickle / unpickle duration objects."""
    from isodate.duration import Duration

    dur = Duration()
    failed = []
    for proto in range(0, pickle.HIGHEST_PROTOCOL + 1):
        try:
            pikl = pickle.dumps(dur, proto)
            if dur != pickle.loads(pikl):
                raise Exception("not equal")
        except Exception as e:
            failed.append("pickle proto %d failed (%s)" % (proto, repr(e)))
    assert len(failed) == 0, "pickle protos failed: %s" % str(failed)


def test_pickle_utc():
    """isodate.UTC objects remain the same after pickling."""
    assert isodate.UTC is pickle.loads(pickle.dumps(isodate.UTC))



"""
This modules provides a method to parse an ISO 8601:2004 time string to a
Python datetime.time instance.

It supports all basic and extended formats including time zone specifications
as described in the ISO standard.
"""

import re
from datetime import time
from decimal import ROUND_FLOOR, Decimal

from isodate.isoerror import ISO8601Error
from isodate.isostrf import TIME_EXT_COMPLETE, TZ_EXT, strftime
from isodate.isotzinfo import TZ_REGEX, build_tzinfo

TIME_REGEX_CACHE = []
# used to cache regular expressions to parse ISO time strings.


def build_time_regexps():
    """
    Build regular expressions to parse ISO time string.

    The regular expressions are compiled and stored in TIME_REGEX_CACHE
    for later reuse.
    """
    if not TIME_REGEX_CACHE:
        # ISO 8601 time representations allow decimal fractions on least
        #    significant time component. Command and Full Stop are both valid
        #    fraction separators.
        #    The letter 'T' is allowed as time designator in front of a time
        #    expression.
        #    Immediately after a time expression, a time zone definition is
        #      allowed.
        #    a TZ may be missing (local time), be a 'Z' for UTC or a string of
        #    +-hh:mm where the ':mm' part can be skipped.
        # TZ information patterns:
        #    ''
        #    Z
        #    +-hh:mm
        #    +-hhmm
        #    +-hh =>
        #    isotzinfo.TZ_REGEX
        def add_re(regex_text):
            TIME_REGEX_CACHE.append(re.compile(r"\A" + regex_text + TZ_REGEX + r"\Z"))

        # 1. complete time:
        #    hh:mm:ss.ss ... extended format
        add_re(
            r"T?(?P<hour>[0-9]{2}):"
            r"(?P<minute>[0-9]{2}):"
            r"(?P<second>[0-9]{2}"
            r"([,.][0-9]+)?)"
        )
        #    hhmmss.ss ... basic format
        add_re(
            r"T?(?P<hour>[0-9]{2})"
            r"(?P<minute>[0-9]{2})"
            r"(?P<second>[0-9]{2}"
            r"([,.][0-9]+)?)"
        )
        # 2. reduced accuracy:
        #    hh:mm.mm ... extended format
        add_re(r"T?(?P<hour>[0-9]{2}):" r"(?P<minute>[0-9]{2}" r"([,.][0-9]+)?)")
        #    hhmm.mm ... basic format
        add_re(r"T?(?P<hour>[0-9]{2})" r"(?P<minute>[0-9]{2}" r"([,.][0-9]+)?)")
        #    hh.hh ... basic format
        add_re(r"T?(?P<hour>[0-9]{2}" r"([,.][0-9]+)?)")
    return TIME_REGEX_CACHE


def parse_time(timestring):
    """
    Parses ISO 8601 times into datetime.time objects.

    Following ISO 8601 formats are supported:
      (as decimal separator a ',' or a '.' is allowed)
      hhmmss.ssTZD    basic complete time
      hh:mm:ss.ssTZD  extended complete time
      hhmm.mmTZD      basic reduced accuracy time
      hh:mm.mmTZD     extended reduced accuracy time
      hh.hhTZD        basic reduced accuracy time
    TZD is the time zone designator which can be in the following format:
              no designator indicates local time zone
      Z       UTC
      +-hhmm  basic hours and minutes
      +-hh:mm extended hours and minutes
      +-hh    hours
    """
    isotimes = build_time_regexps()
    for pattern in isotimes:
        match = pattern.match(timestring)
        if match:
            groups = match.groupdict()
            for key, value in groups.items():
                if value is not None:
                    groups[key] = value.replace(",", ".")
            tzinfo = build_tzinfo(
                groups["tzname"],
                groups["tzsign"],
                int(groups["tzhour"] or 0),
                int(groups["tzmin"] or 0),
            )
            if "second" in groups:
                second = Decimal(groups["second"]).quantize(
                    Decimal(".000001"), rounding=ROUND_FLOOR
                )
                microsecond = (second - int(second)) * int(1e6)
                # int(...) ... no rounding
                # to_integral() ... rounding
                return time(
                    int(groups["hour"]),
                    int(groups["minute"]),
                    int(second),
                    int(microsecond.to_integral()),
                    tzinfo,
                )
            if "minute" in groups:
                minute = Decimal(groups["minute"])
                second = Decimal((minute - int(minute)) * 60).quantize(
                    Decimal(".000001"), rounding=ROUND_FLOOR
                )
                microsecond = (second - int(second)) * int(1e6)
                return time(
                    int(groups["hour"]),
                    int(minute),
                    int(second),
                    int(microsecond.to_integral()),
                    tzinfo,
                )
            else:
                microsecond, second, minute = 0, 0, 0
            hour = Decimal(groups["hour"])
            minute = (hour - int(hour)) * 60
            second = (minute - int(minute)) * 60
            microsecond = (second - int(second)) * int(1e6)
            return time(
                int(hour),
                int(minute),
                int(second),
                int(microsecond.to_integral()),
                tzinfo,
            )
    raise ISO8601Error("Unrecognised ISO 8601 time format: %r" % timestring)


def time_isoformat(ttime, format=TIME_EXT_COMPLETE + TZ_EXT):
    """
    Format time strings.

    This method is just a wrapper around isodate.isostrf.strftime and uses
    Time-Extended-Complete with extended time zone as default format.
    """
    return strftime(ttime, format)



"""
Test cases for the isodate module.
"""

from datetime import date

import pytest

from isodate import (
    DATE_BAS_COMPLETE,
    DATE_BAS_MONTH,
    DATE_BAS_ORD_COMPLETE,
    DATE_BAS_WEEK,
    DATE_BAS_WEEK_COMPLETE,
    DATE_CENTURY,
    DATE_EXT_COMPLETE,
    DATE_EXT_MONTH,
    DATE_EXT_ORD_COMPLETE,
    DATE_EXT_WEEK,
    DATE_EXT_WEEK_COMPLETE,
    DATE_YEAR,
    ISO8601Error,
    date_isoformat,
    parse_date,
)

# the following list contains tuples of ISO date strings and the expected
# result from the parse_date method. A result of None means an ISO8601Error
# is expected. The test cases are grouped into dates with 4 digit years
# and 6 digit years.
TEST_CASES = {
    # yeardigits = 4
    (4, "19", date(1901, 1, 1), DATE_CENTURY),
    (4, "1985", date(1985, 1, 1), DATE_YEAR),
    (4, "1985-04", date(1985, 4, 1), DATE_EXT_MONTH),
    (4, "198504", date(1985, 4, 1), DATE_BAS_MONTH),
    (4, "1985-04-12", date(1985, 4, 12), DATE_EXT_COMPLETE),
    (4, "19850412", date(1985, 4, 12), DATE_BAS_COMPLETE),
    (4, "1985102", date(1985, 4, 12), DATE_BAS_ORD_COMPLETE),
    (4, "1985-102", date(1985, 4, 12), DATE_EXT_ORD_COMPLETE),
    (4, "1985W155", date(1985, 4, 12), DATE_BAS_WEEK_COMPLETE),
    (4, "1985-W15-5", date(1985, 4, 12), DATE_EXT_WEEK_COMPLETE),
    (4, "1985W15", date(1985, 4, 8), DATE_BAS_WEEK),
    (4, "1985-W15", date(1985, 4, 8), DATE_EXT_WEEK),
    (4, "1989-W15", date(1989, 4, 10), DATE_EXT_WEEK),
    (4, "1989-W15-5", date(1989, 4, 14), DATE_EXT_WEEK_COMPLETE),
    (4, "1-W1-1", None, DATE_BAS_WEEK_COMPLETE),
    # yeardigits = 6
    (6, "+0019", date(1901, 1, 1), DATE_CENTURY),
    (6, "+001985", date(1985, 1, 1), DATE_YEAR),
    (6, "+001985-04", date(1985, 4, 1), DATE_EXT_MONTH),
    (6, "+001985-04-12", date(1985, 4, 12), DATE_EXT_COMPLETE),
    (6, "+0019850412", date(1985, 4, 12), DATE_BAS_COMPLETE),
    (6, "+001985102", date(1985, 4, 12), DATE_BAS_ORD_COMPLETE),
    (6, "+001985-102", date(1985, 4, 12), DATE_EXT_ORD_COMPLETE),
    (6, "+001985W155", date(1985, 4, 12), DATE_BAS_WEEK_COMPLETE),
    (6, "+001985-W15-5", date(1985, 4, 12), DATE_EXT_WEEK_COMPLETE),
    (6, "+001985W15", date(1985, 4, 8), DATE_BAS_WEEK),
    (6, "+001985-W15", date(1985, 4, 8), DATE_EXT_WEEK),
}


@pytest.mark.parametrize("yeardigits,datestring,expected,_", TEST_CASES)
def test_parse(yeardigits, datestring, expected, _):
    if expected is None:
        with pytest.raises(ISO8601Error):
            parse_date(datestring, yeardigits)
    else:
        result = parse_date(datestring, yeardigits)
        assert result == expected


@pytest.mark.parametrize("yeardigits, datestring, expected, format", TEST_CASES)
def test_format(yeardigits, datestring, expected, format):
    """
    Take date object and create ISO string from it.
    This is the reverse test to test_parse.
    """
    if expected is None:
        with pytest.raises(AttributeError):
            date_isoformat(expected, format, yeardigits)
    else:
        assert date_isoformat(expected, format, yeardigits) == datestring



"""
Test cases for the isotime module.
"""

from datetime import time

import pytest

from isodate import (
    TIME_BAS_COMPLETE,
    TIME_BAS_MINUTE,
    TIME_EXT_COMPLETE,
    TIME_EXT_MINUTE,
    TIME_HOUR,
    TZ_BAS,
    TZ_EXT,
    TZ_HOUR,
    UTC,
    FixedOffset,
    ISO8601Error,
    parse_time,
    time_isoformat,
)

# the following list contains tuples of ISO time strings and the expected
# result from the parse_time method. A result of None means an ISO8601Error
# is expected.
TEST_CASES = [
    ("232050", time(23, 20, 50), TIME_BAS_COMPLETE + TZ_BAS),
    ("23:20:50", time(23, 20, 50), TIME_EXT_COMPLETE + TZ_EXT),
    ("2320", time(23, 20), TIME_BAS_MINUTE),
    ("23:20", time(23, 20), TIME_EXT_MINUTE),
    ("23", time(23), TIME_HOUR),
    ("232050,5", time(23, 20, 50, 500000), None),
    ("23:20:50.5", time(23, 20, 50, 500000), None),
    # test precision
    ("15:33:42.123456", time(15, 33, 42, 123456), None),
    ("15:33:42.1234564", time(15, 33, 42, 123456), None),
    ("15:33:42.1234557", time(15, 33, 42, 123455), None),
    (
        "10:59:59.9999999Z",
        time(10, 59, 59, 999999, tzinfo=UTC),
        None,
    ),  # TIME_EXT_COMPLETE + TZ_EXT),
    ("2320,8", time(23, 20, 48), None),
    ("23:20,8", time(23, 20, 48), None),
    ("23,3", time(23, 18), None),
    ("232030Z", time(23, 20, 30, tzinfo=UTC), TIME_BAS_COMPLETE + TZ_BAS),
    ("2320Z", time(23, 20, tzinfo=UTC), TIME_BAS_MINUTE + TZ_BAS),
    ("23Z", time(23, tzinfo=UTC), TIME_HOUR + TZ_BAS),
    ("23:20:30Z", time(23, 20, 30, tzinfo=UTC), TIME_EXT_COMPLETE + TZ_EXT),
    ("23:20Z", time(23, 20, tzinfo=UTC), TIME_EXT_MINUTE + TZ_EXT),
    (
        "152746+0100",
        time(15, 27, 46, tzinfo=FixedOffset(1, 0, "+0100")),
        TIME_BAS_COMPLETE + TZ_BAS,
    ),
    (
        "152746-0500",
        time(15, 27, 46, tzinfo=FixedOffset(-5, 0, "-0500")),
        TIME_BAS_COMPLETE + TZ_BAS,
    ),
    (
        "152746+01",
        time(15, 27, 46, tzinfo=FixedOffset(1, 0, "+01:00")),
        TIME_BAS_COMPLETE + TZ_HOUR,
    ),
    (
        "152746-05",
        time(15, 27, 46, tzinfo=FixedOffset(-5, -0, "-05:00")),
        TIME_BAS_COMPLETE + TZ_HOUR,
    ),
    (
        "15:27:46+01:00",
        time(15, 27, 46, tzinfo=FixedOffset(1, 0, "+01:00")),
        TIME_EXT_COMPLETE + TZ_EXT,
    ),
    (
        "15:27:46-05:00",
        time(15, 27, 46, tzinfo=FixedOffset(-5, -0, "-05:00")),
        TIME_EXT_COMPLETE + TZ_EXT,
    ),
    (
        "15:27:46+01",
        time(15, 27, 46, tzinfo=FixedOffset(1, 0, "+01:00")),
        TIME_EXT_COMPLETE + TZ_HOUR,
    ),
    (
        "15:27:46-05",
        time(15, 27, 46, tzinfo=FixedOffset(-5, -0, "-05:00")),
        TIME_EXT_COMPLETE + TZ_HOUR,
    ),
    (
        "15:27:46-05:30",
        time(15, 27, 46, tzinfo=FixedOffset(-5, -30, "-05:30")),
        TIME_EXT_COMPLETE + TZ_EXT,
    ),
    (
        "15:27:46-0545",
        time(15, 27, 46, tzinfo=FixedOffset(-5, -45, "-0545")),
        TIME_EXT_COMPLETE + TZ_BAS,
    ),
    ("1:17:30", None, TIME_EXT_COMPLETE),
]


@pytest.mark.parametrize("timestring, expectation, format", TEST_CASES)
def test_parse(timestring, expectation, format):
    """
    Parse an ISO time string and compare it to the expected value.
    """
    if expectation is None:
        with pytest.raises(ISO8601Error):
            parse_time(timestring)
    else:
        assert parse_time(timestring) == expectation


@pytest.mark.parametrize("timestring, expectation, format", TEST_CASES)
def test_format(timestring, expectation, format):
    """
    Take time object and create ISO string from it.
    This is the reverse test to test_parse.
    """
    if expectation is None:
        with pytest.raises(AttributeError):
            time_isoformat(expectation, format)
    elif format is not None:
        assert time_isoformat(expectation, format) == timestring



"""Test cases for the isodate module."""

import time
from datetime import datetime, timedelta

import pytest

from isodate import DT_EXT_COMPLETE, LOCAL, strftime, tzinfo

TEST_CASES = (
    (
        datetime(2012, 12, 25, 13, 30, 0, 0, LOCAL),
        DT_EXT_COMPLETE,
        "2012-12-25T13:30:00+10:00",
    ),
    # DST ON
    (
        datetime(1999, 12, 25, 13, 30, 0, 0, LOCAL),
        DT_EXT_COMPLETE,
        "1999-12-25T13:30:00+11:00",
    ),
    # microseconds
    (
        datetime(2012, 10, 12, 8, 29, 46, 69178),
        "%Y-%m-%dT%H:%M:%S.%f",
        "2012-10-12T08:29:46.069178",
    ),
    (
        datetime(2012, 10, 12, 8, 29, 46, 691780),
        "%Y-%m-%dT%H:%M:%S.%f",
        "2012-10-12T08:29:46.691780",
    ),
)


@pytest.fixture
def tz_patch(monkeypatch):
    # local time zone mock function
    localtime_orig = time.localtime

    def localtime_mock(secs):
        """Mock time to fixed date.

        Mock time.localtime so that it always returns a time_struct with tm_dst=1
        """
        tt = localtime_orig(secs)
        # before 2000 everything is dst, after 2000 no dst.
        if tt.tm_year < 2000:
            dst = 1
        else:
            dst = 0
        tt = (
            tt.tm_year,
            tt.tm_mon,
            tt.tm_mday,
            tt.tm_hour,
            tt.tm_min,
            tt.tm_sec,
            tt.tm_wday,
            tt.tm_yday,
            dst,
        )
        return time.struct_time(tt)

    monkeypatch.setattr(time, "localtime", localtime_mock)
    # assume LOC = +10:00
    monkeypatch.setattr(tzinfo, "STDOFFSET", timedelta(seconds=36000))
    # assume DST = +11:00
    monkeypatch.setattr(tzinfo, "DSTOFFSET", timedelta(seconds=39600))
    monkeypatch.setattr(tzinfo, "DSTDIFF", tzinfo.DSTOFFSET - tzinfo.STDOFFSET)


@pytest.mark.parametrize("dt, format, expectation", TEST_CASES)
def test_format(tz_patch, dt, format, expectation):
    """Take date object and create ISO string from it.

    This is the reverse test to test_parse.
    """
    if expectation is None:
        with pytest.raises(AttributeError):
            strftime(dt, format)
    else:
        assert strftime(dt, format) == expectation
